/*
  Liquid War 6 is a unique multiplayer wargame.
  Copyright (C)  2005, 2006, 2007, 2008, 2009  Christian Mauduit <ufoot@ufoot.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  

  Liquid War 6 homepage : http://www.gnu.org/software/liquidwar6/
  Contact author        : ufoot@ufoot.org
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <GL/glu.h>

#include "../../../gfx.h"
#include "gl-flat.h"
#include "gl-flat-internal.h"

static void
set_map_rules (mod_gl_utils_context_t * utils_context,
	       _mod_gl_view_flat_context_t * flat_context)
{
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
}

static void
display_texture_ex (mod_gl_utils_context_t * utils_context,
		    _mod_gl_view_flat_context_t * flat_context,
		    GLuint texture, float x1, float y1, float x2, float y2,
		    float scale_w, float scale_h)
{
  glBindTexture (GL_TEXTURE_2D, texture);
  set_map_rules (utils_context, flat_context);

  glMatrixMode (GL_TEXTURE);
  glPushMatrix ();
  glLoadIdentity ();
  glBegin (GL_QUADS);
  glTexCoord2d (0.0f, 0.0f);
  glVertex3f (x1, y1, 0.0f);	// top left
  glTexCoord2d (scale_w, 0.0f);
  glVertex3f (x2, y1, 0.0f);	// top right
  glTexCoord2d (scale_w, scale_h);
  glVertex3f (x2, y2, 0.0f);	// bottom right
  glTexCoord2d (0.0f, scale_h);
  glVertex3f (x1, y2, 0.0f);	// bottom left
  glEnd ();

  glMatrixMode (GL_TEXTURE);
  glPopMatrix ();
}

static void
display_texture (mod_gl_utils_context_t * utils_context,
		 _mod_gl_view_flat_context_t * flat_context,
		 int texture_w, int texture_h, int shape_w, int shape_h,
		 GLuint texture, int center_x, int center_y,
		 lw6gui_look_t * look)
{
  float x1, y1, x2, y2, scale_w, scale_h;

  scale_w = mod_gl_utils_texture_scale (texture_w);
  scale_h = mod_gl_utils_texture_scale (texture_h);

  mod_gl_utils_viewport_calc_map_area (utils_context, &x1, &y1, &x2, &y2,
				       center_x, center_y, shape_w, shape_h,
				       look);

  display_texture_ex (utils_context,
		      flat_context,
		      texture, x1, y1, x2, y2, scale_w, scale_h);
}

static void
display_texture_array (mod_gl_utils_context_t * utils_context,
		       _mod_gl_view_flat_context_t * flat_context,
		       int texture_array_w,
		       int texture_array_h,
		       int shape_w,
		       int shape_h,
		       mod_gl_utils_texture_array_t * texture_array,
		       int center_x, int center_y, lw6gui_look_t * look)
{
  float x1, y1, x2, y2, w, h, coef_x, coef_y;
  float scale_w, scale_h;
  int n_x, n_y;
  float *array_x = NULL;
  float *array_y = NULL;
  int view_x, view_y, view_w, view_h;
  float min_x, min_y, max_x, max_y;

  mod_gl_utils_viewport_calc_map_area (utils_context, &x1, &y1, &x2, &y2,
				       center_x, center_y,
				       shape_w, shape_h, look);
  w = x2 - x1;
  h = y2 - y1;
  coef_x = w / texture_array_w;
  coef_y = h / texture_array_h;

  mod_gl_utils_viewport_calc_view_area (utils_context, &view_x, &view_y,
					&view_w, &view_h);
  min_x = view_x;
  min_y = view_y;
  max_x = view_x + view_w;
  max_y = view_y + view_h;

  array_x = LW6SYS_MALLOC ((texture_array->layout.n_w + 1) * sizeof (float));
  array_y = LW6SYS_MALLOC ((texture_array->layout.n_h + 1) * sizeof (float));

  if (array_x && array_y)
    {
      if (texture_array->layout.n_w > 0)
	{
	  for (n_x = 0; n_x < texture_array->layout.n_w; ++n_x)
	    {
	      array_x[n_x] = x1 + texture_array->layout.x0[n_x] * coef_x;
	    }
	  array_x[texture_array->layout.n_w] =
	    array_x[texture_array->layout.n_w - 1] +
	    texture_array->layout.w[texture_array->layout.n_w - 1] * coef_x;
	}

      if (texture_array->layout.n_h > 0)
	{
	  for (n_y = 0; n_y < texture_array->layout.n_h; ++n_y)
	    {
	      array_y[n_y] = y1 + texture_array->layout.y0[n_y] * coef_y;
	    }
	  array_y[texture_array->layout.n_h] =
	    array_y[texture_array->layout.n_h - 1] +
	    texture_array->layout.h[texture_array->layout.n_h - 1] * coef_y;
	}

      for (n_y = 0; n_y < texture_array->layout.n_h; ++n_y)
	{
	  for (n_x = 0; n_x < texture_array->layout.n_w; ++n_x)
	    {
	      if (array_x[n_x] > max_x ||
		  array_y[n_y] > max_y ||
		  array_x[n_x + 1] < min_x || array_y[n_y + 1] < min_y)
		{
		  // skip texture, it's outside our screen...
		}
	      else
		{
		  scale_w =
		    mod_gl_utils_texture_scale (texture_array->layout.w[n_x]);
		  scale_h =
		    mod_gl_utils_texture_scale (texture_array->layout.h[n_y]);
		  display_texture_ex (utils_context,
				      flat_context,
				      mod_gl_utils_get_texture_array_texture
				      (texture_array, n_x, n_y),
				      array_x[n_x],
				      array_y[n_y],
				      array_x[n_x + 1],
				      array_y[n_y + 1], scale_w, scale_h);
		}
	    }
	}
      LW6SYS_FREE (array_x);
      LW6SYS_FREE (array_y);
    }
}

static void
display_map_preview (mod_gl_utils_context_t * utils_context,
		     _mod_gl_view_flat_context_t * flat_context,
		     lw6gui_look_t * look, lw6map_level_t * level,
		     int center_x, int center_y)
{
  display_texture_array (utils_context, flat_context,
			 flat_context->game_context.map.level->texture.w,
			 flat_context->game_context.map.level->texture.h,
			 flat_context->game_context.map.level->body.shape.w,
			 flat_context->game_context.map.level->body.shape.h,
			 &(flat_context->game_context.map.map_texture_array),
			 center_x, center_y, look);
}

void
_mod_gl_view_flat_display_preview (mod_gl_utils_context_t * utils_context,
				   _mod_gl_view_flat_context_t *
				   flat_context, lw6gui_look_t * look,
				   lw6map_level_t * level)
{
  if (utils_context && flat_context && level && level->texture.data)
    {
      _mod_gl_view_flat_game_context_update_map (utils_context,
						 flat_context,
						 &(flat_context->
						   game_context.map), level);

      mod_gl_utils_set_render_mode_2d_blend (utils_context);

      glColor3f (1.0, 1.0, 1.0);
      glEnable (GL_TEXTURE_2D);	// for texture

      display_map_preview (utils_context, flat_context, look, level,
			   level->body.shape.w / 2, level->body.shape.h / 2);
    }
}

void
mod_gl_view_flat_display_preview (mod_gl_utils_context_t * utils_context,
				  void *flat_context, lw6gui_look_t * look,
				  lw6map_level_t * level)
{
  _mod_gl_view_flat_display_preview (utils_context,
				     (_mod_gl_view_flat_context_t *)
				     flat_context, look, level);
}

static void
display_map_zones (mod_gl_utils_context_t * utils_context,
		   _mod_gl_view_flat_context_t * flat_context,
		   lw6gui_look_t * look, lw6map_level_t * level,
		   lw6ker_map_struct_t * map_struct)
{
  GLuint texture = 0;
  SDL_Surface *surface = NULL;

  surface = mod_gl_utils_create_zones_surface (utils_context, map_struct);
  if (surface)
    {
      texture = mod_gl_utils_surface2texture (utils_context, surface);

      if (texture)
	{
	  display_texture (utils_context, flat_context, level->texture.w,
			   level->texture.h, map_struct->shape.w,
			   map_struct->shape.h, texture,
			   map_struct->shape.w / 2, map_struct->shape.h / 2,
			   look);

	  mod_gl_utils_schedule_delete_texture (utils_context, texture);
	}

      mod_gl_utils_delete_surface (utils_context, surface);
    }
}

void
_mod_gl_view_flat_display_zones (mod_gl_utils_context_t * utils_context,
				 _mod_gl_view_flat_context_t *
				 flat_context, lw6gui_look_t * look,
				 lw6map_level_t * level,
				 lw6ker_map_struct_t * map_struct)
{
  if (utils_context && flat_context && level && map_struct)
    {
      //_mod_gl_set_render_mode_2d (context);
      mod_gl_utils_set_render_mode_2d_blend (utils_context);

      glColor3f (1.0, 1.0, 1.0);
      glEnable (GL_TEXTURE_2D);	// for texture

      display_map_zones (utils_context, flat_context, look, level,
			 map_struct);
    }
}

void
mod_gl_view_flat_display_zones (mod_gl_utils_context_t * utils_context,
				void *flat_context, lw6gui_look_t * look,
				lw6map_level_t * level,
				lw6ker_map_struct_t * map_struct)
{
  _mod_gl_view_flat_display_zones (utils_context,
				   (_mod_gl_view_flat_context_t *)
				   flat_context, look, level, map_struct);
}

static void
display_map_gradient (mod_gl_utils_context_t * utils_context,
		      _mod_gl_view_flat_context_t * flat_context,
		      lw6gui_look_t * look,
		      lw6map_level_t * level,
		      lw6ker_map_state_t * map_state, int team_id,
		      int layer_id)
{
  GLuint texture = 0;
  SDL_Surface *surface = NULL;

  surface =
    mod_gl_utils_create_gradient_surface (utils_context, map_state, team_id,
					  layer_id);
  if (surface)
    {
      texture = mod_gl_utils_surface2texture (utils_context, surface);

      if (texture)
	{
	  display_texture (utils_context, flat_context, level->texture.w,
			   level->texture.h, map_state->shape.w,
			   map_state->shape.h, texture,
			   map_state->shape.w / 2, map_state->shape.h / 2,
			   look);

	  mod_gl_utils_schedule_delete_texture (utils_context, texture);
	}

      mod_gl_utils_delete_surface (utils_context, surface);
    }
}

void
_mod_gl_view_flat_display_gradient (mod_gl_utils_context_t * utils_context,
				    _mod_gl_view_flat_context_t *
				    flat_context, lw6gui_look_t * look,
				    lw6map_level_t * level,
				    lw6ker_map_state_t * map_state,
				    int team_id, int layer_id)
{
  if (utils_context && flat_context && map_state)
    {
      mod_gl_utils_set_render_mode_2d_blend (utils_context);

      glColor3f (1.0, 1.0, 1.0);
      glEnable (GL_TEXTURE_2D);	// for texture

      display_map_gradient (utils_context, flat_context, look, level,
			    map_state, team_id, layer_id);
    }
}

void
mod_gl_view_flat_display_gradient (mod_gl_utils_context_t * utils_context,
				   void *flat_context,
				   lw6gui_look_t * look,
				   lw6map_level_t * level,
				   lw6ker_map_state_t * map_state,
				   int team_id, int layer_id)
{
  _mod_gl_view_flat_display_gradient (utils_context,
				      (_mod_gl_view_flat_context_t *)
				      flat_context, look, level, map_state,
				      team_id, layer_id);
}

/*
 * Game_state is not passed as args but retrived
 * from flat_context struct, not to give the illusion that
 * any game_state could be displayed. One *needs* to call
 * game_context_init first.
 */
static void
display_armies (mod_gl_utils_context_t * utils_context,
		_mod_gl_view_flat_context_t *
		flat_context, int center_x, int center_y,
		lw6gui_look_t * look)
{
  mod_gl_utils_update_game_texture_array (utils_context,
					  &(flat_context->game_context.
					    armies.armies_surface_array),
					  &(flat_context->game_context.armies.
					    armies_texture_array),
					  flat_context->game_context.
					  armies.game_state, look);
  display_texture_array (utils_context, flat_context,
			 flat_context->game_context.armies.game_state->
			 map_state.shape.w,
			 flat_context->game_context.armies.game_state->
			 map_state.shape.h,
			 flat_context->game_context.armies.game_state->
			 map_state.shape.w,
			 flat_context->game_context.armies.game_state->
			 map_state.shape.h,
			 &(flat_context->game_context.armies.
			   armies_texture_array), center_x, center_y, look);
}

/*
 * Game_state is not passed as args but retrived
 * from flat_context struct, not to give the illusion that
 * any game_state could be displayed. One *needs* to call
 * game_context_init first.
 */
static void
display_game (mod_gl_utils_context_t * utils_context,
	      _mod_gl_view_flat_context_t *
	      flat_context, lw6gui_look_t * look,
	      lw6ker_game_state_t * game_state, u_int16_t main_cursor_id)
{
  int center_x = 0;
  int center_y = 0;

  lw6sys_log (LW6SYS_LOG_DEBUG, _("display game"));
  lw6ker_game_state_get_cursor_info (game_state, main_cursor_id,
				     NULL, NULL, NULL, &center_x, &center_y);
  display_map_preview (utils_context, flat_context, look,
		       game_state->game_struct->level, center_x, center_y);
  display_armies (utils_context, flat_context, center_x, center_y, look);
}

void
_mod_gl_view_flat_display_game (mod_gl_utils_context_t * utils_context,
				_mod_gl_view_flat_context_t *
				flat_context,
				lw6gui_look_t * look,
				lw6ker_game_state_t * game_state,
				u_int16_t main_cursor_id)
{
  if (utils_context && flat_context && game_state)
    {
      lw6map_level_t *level;

      level = game_state->game_struct->level;

      _mod_gl_view_flat_game_context_update (utils_context,
					     flat_context,
					     &flat_context->game_context,
					     look, level, game_state);

      mod_gl_utils_set_render_mode_2d_blend (utils_context);

      glColor3f (1.0, 1.0, 1.0);
      glEnable (GL_TEXTURE_2D);	// for texture

      display_game (utils_context, flat_context, look, game_state,
		    main_cursor_id);
    }
}

void
mod_gl_view_flat_display_game (mod_gl_utils_context_t * utils_context,
			       void *flat_context,
			       lw6gui_look_t * look,
			       lw6ker_game_state_t * game_state,
			       u_int16_t main_cursor_id)
{
  _mod_gl_view_flat_display_game (utils_context,
				  (_mod_gl_view_flat_context_t *)
				  flat_context, look, game_state,
				  main_cursor_id);
}

static void
display_cursor_texture (mod_gl_utils_context_t * utils_context,
			_mod_gl_view_flat_context_t * flat_context,
			float x, float y, float w, float h, GLuint texture)
{
  display_texture_ex (utils_context, flat_context,
		      texture, x - w / 2.0f, y - h / 2.0f, x + w / 2.0f,
		      y + h / 2.0f, 1.0f, 1.0f);
}

static void
display_cursor (mod_gl_utils_context_t * utils_context,
		_mod_gl_view_flat_context_t * flat_context,
		lw6gui_look_t * look,
		int i, int x, int y, int w, int h, int center_x, int center_y,
		int even_odd)
{

  float cursor_x = 0.0f;
  float cursor_y = 0.0f;
  float cursor_w = 0.0f;
  float cursor_h = 0.0f;

  mod_gl_utils_viewport_calc_map_pos (utils_context, &cursor_x, &cursor_y,
				      x, y, center_x, center_y, w, h, look);
  cursor_w = cursor_h =
    (utils_context->screen.width +
     utils_context->screen.height) * flat_context->const_data.cursor_size *
    look->style.cursor_size;
  lw6sys_log (LW6SYS_LOG_DEBUG,
	      _("display cursor %d %0.1f , %0.1f - %0.1f x %0.1f"), i,
	      cursor_x, cursor_y, cursor_w, cursor_h);
  display_cursor_texture (utils_context, flat_context, cursor_x, cursor_y,
			  cursor_w, cursor_h, even_odd ?
			  flat_context->cursors_context.
			  cursor[i].texture_even : flat_context->
			  cursors_context.cursor[i].texture_odd);
}

static void
display_cursors (mod_gl_utils_context_t * utils_context,
		 _mod_gl_view_flat_context_t *
		 flat_context,
		 lw6gui_look_t * look, lw6ker_game_state_t * game_state,
		 u_int16_t main_cursor_id)
{
  int i;
  lw6ker_cursor_t *cursor;
  int blink_state;
  int center_x = 0;
  int center_y = 0;

  lw6sys_log (LW6SYS_LOG_DEBUG, _("display cursors"));
  blink_state =
    mod_gl_utils_get_ticks (utils_context) /
    (flat_context->const_data.cursor_blink_period / 2) % 2;
  lw6ker_game_state_get_cursor_info (game_state, main_cursor_id,
				     NULL, NULL, NULL, &center_x, &center_y);
  for (i = 0; i < LW6MAP_MAX_NB_CURSORS; ++i)
    {
      cursor = &(game_state->map_state.cursor_array.cursors[i]);
      if (cursor->enabled)
	{
	  display_cursor (utils_context, flat_context, look,
			  i, cursor->pos.x, cursor->pos.y,
			  game_state->map_state.shape.w,
			  game_state->map_state.shape.h, center_x, center_y,
			  blink_state);
	}
    }
}

void
_mod_gl_view_flat_display_cursors (mod_gl_utils_context_t * utils_context,
				   _mod_gl_view_flat_context_t *
				   flat_context,
				   lw6gui_look_t * look,
				   lw6ker_game_state_t * game_state,
				   u_int16_t main_cursor_id)
{
  if (utils_context && flat_context && game_state)
    {
      _mod_gl_view_flat_cursors_context_update (utils_context,
						&flat_context->
						cursors_context,
						&flat_context->const_data,
						look, game_state);

      mod_gl_utils_set_render_mode_2d_blend (utils_context);

      glColor3f (1.0, 1.0, 1.0);
      glEnable (GL_TEXTURE_2D);	// for texture

      display_cursors (utils_context, flat_context, look, game_state,
		       main_cursor_id);
    }
}

void
mod_gl_view_flat_display_cursors (mod_gl_utils_context_t *
				  utils_context,
				  void *flat_context,
				  lw6gui_look_t * look,
				  lw6ker_game_state_t * game_state,
				  u_int16_t main_cursor_id)
{
  _mod_gl_view_flat_display_cursors (utils_context,
				     (_mod_gl_view_flat_context_t *)
				     flat_context, look, game_state,
				     main_cursor_id);
}
